import curifactory.utils
from curifactory import docker

from pytest_mock import mocker  # noqa: F401 -- flake8 doesn't see it's used as fixture


def test_build_docker(mocker):  # noqa: F811 -- mocker has to be passed in as fixture
    """This is mostly just to make sure the docker command isn't changed from what we
    expect without a notification."""

    mocker.patch("curifactory.utils.run_command")
    docker.build_docker("test_name", "./test_cache_folder", "some_version_string")

    curifactory.utils.run_command.assert_called_once_with(
        "docker",
        "build",
        "-f",
        "docker/dockerfile",
        "--tag",
        "test_name:some_version_string",
        "--tag",
        "test_name",
        "--build-arg",
        "run_folder=./test_cache_folder",
        ".",
        "--progress",
        "tty",
    )



"""Contains the parent dataclass ExperimentArgs, containing run-specific config params."""

from dataclasses import dataclass


@dataclass
class ExperimentArgs:
    """Base arguments class, for handling naming and hashing.

    In any given repo, this class should be extended to contain any needed
    local configuration.

    .. note::

        Extending with a :code:`@dataclass` is recommended to make it syntactically
        easier to read and define.

    Example:
        .. code-block:: python

            from dataclasses import dataclass
            from curifactory import ExperimentArgs

            @dataclass
            class Args(ExperimentArgs):
                some_parameter: int = 0
                # ...
    """

    name: str = "UNNAMED"
    """Argument set name. This can be used to easily distinguish/refer to specific
        configurations in aggregate stages. This should be unique for every args instance."""
    hash: str = None
    """Curifactory automatically fills this, but it can be overriden if you need to use
        very specific cache naming. (Should not normally be necessary.)"""
    overwrite: bool = False
    """Whether to overwrite pre-cached values. Curifactory automatically sets this based
        on command line flags."""



"""Contains the relevant classes for experiment state, namely the artifact manager and
record classes, as well as experiment store/parameter store management."""

from datetime import datetime
import json
import logging
import multiprocessing as mp
import os
from socket import gethostname
import sys
from typing import List, Dict

from curifactory import utils, reporting
from curifactory.reporting import Reportable
from curifactory.record import Record
from curifactory.store import ManagerStore


class ArtifactManager:
    """This class manages the records, metadata, and options for an experiment run.

    Every experiment run needs an instance of this.

    Args:
        experiment_name (str): The name of the experiment. This is used as a prefix for
            most relevant information and affects caching paths. If a manager is being
            used primarily from jupyter notebooks, you could for instance set this to
            be the name of the notebook.
        store_entire_run (bool): Store/copy environment info, log, output report, and
            all cached files in a run-specific folder (this is a --store-full run.)
        dry (bool): Setting dry to true will suppress saving any files (including logs), and will
            not update parameter stores. (It should have no effect on any files.)
        dry_cache (bool): Setting this to true only suppresses saving cache files. This is recommended
            if you're running with a cache_dir_override for some previous --store-full run, so you
            don't accidentally overwrite or add new data to the --store-full directory.
        custom_name (str): Instead of using the experiment name to group cached data, use
            this name instead.
        run_line (str): The CLI command used to run the current experiment.
        parallel_lock (multiprocessing.Lock): If this function is called from a multiprocessing
            context, use this lock to help prevent files being written to and read simultaneously.
        parallel_mode (bool): This is handled by the parallel parameter, informing a particular subproc
            run that it is being executed from a parallel run.
        lazy (bool): If true, attempts to set all stage outputs as Lazy objects. Outputs that do not have
            a cacher specified will be given a PickleCacher. Note that objects without a cacher tht do not
            handle pickle serialization correctly may cause errors.
        ignore_lazy (bool): Run the experiment disabling any lazy object caching/keeping everything in memory.
            This can save time when memory is less of an issue.
        notes (str): A git-log-like message to store in the run info for the current run. If this is an
            empty string, query the user for an input string.
        manager_cache_path (str): The path where the parameter registry and experiment
            store data are kept.
        cache_path (str): The path where all intermediate data is cached.
        runs_path (str): The path where full experiment stores (run with :code:`--store-full`)
            are kept.
        logs_path (str): The path where run logs get stored.
        notebooks_path (str): The path where run notebooks are placed.
        reports_path (str): The path where experiment run reports are saved.
        report_css_path (str): The path to a CSS file to copy into each report directory.
        status_override (str): This variable is 'LIVE' by default, but is overriden by the experiment script.
            'LIVE' indicates that this manager is being run from a non-experiment script or interactive terminal.
        suppress_live_log (bool): If true, on live/interactive runs don't automatically spawn a logger.
        live_log_debug (bool): If true, spawn live logger with the DEBUG level.
    """

    def __init__(
        self,
        experiment_name: str = None,
        store_entire_run: bool = False,
        dry: bool = False,
        dry_cache: bool = False,
        custom_name: str = None,
        run_line: str = "",
        parallel_lock: mp.Lock = None,
        parallel_mode: bool = False,
        lazy: bool = False,
        ignore_lazy: bool = False,
        notes: str = None,
        manager_cache_path: str = None,
        cache_path: str = None,
        runs_path: str = None,
        logs_path: str = None,
        notebooks_path: str = None,
        reports_path: str = None,
        report_css_path: str = None,
        status_override: str = "LIVE",
        suppress_live_log: bool = False,
        live_log_debug: bool = False,
    ):
        self.current_stage_name = ""
        """The name of the stage currently executing."""

        self.experiment_name = experiment_name
        """The name of the experiment and/or the prefix used for caching."""
        self.run_timestamp = datetime.now()
        """The datetime timestamp for when the manager is initialized (and usually
        also when the experiment starts running.)"""
        self.experiment_run_number = 0
        """The run counter for experiments with the given name."""
        self.experiment_args_file_list = []
        """The list of parameter file names to be used in the experiment."""
        self.experiment_args = {}
        """A dictionary keyed by parameter file names, where the values are arrays of the name of the
        argset and arg hashes for all argsets from that parameter file. (See the 'args' field in the
        metadata blocks in ManagerStore.)"""
        self.git_commit_hash = ""
        """The current commit hash if a git repo is in use."""
        self.pip_freeze = ""
        """The output from a :code:`pip freeze` command."""
        self.conda_env = ""
        """The output from a conda env command (:code:`conda env export --from-history`)."""
        self.os = ""
        """The name of the current OS running curifactory."""
        self.hostname = gethostname()
        """The hostname of the machine this experiment ran on."""

        self.custom_name = custom_name
        """If specified, the name to use for grouping cached data instead of the experiment name."""
        self.notes = notes
        """A notes associated with a session/run to output into the report etc."""

        self.manager_cache_path = manager_cache_path
        """The path where the parameter registry and experiment store data are kept."""
        self.cache_path = cache_path
        """The path where all intermediate data is cached."""
        self.runs_path = runs_path
        """The path where full experiment stores (run with :code:`--store-full`) are kept."""
        self.logs_path = logs_path
        """The path where run logs get stored."""
        self.notebooks_path = notebooks_path
        """The path where run notebooks are placed."""
        self.reports_path = reports_path
        """The path where experiment run reports are saved."""
        self.report_css_path = report_css_path
        """The path to a CSS file to copy into each report directory."""

        self.config = {}
        """The manager configuration loaded from the curifactory config file if present."""

        self.records = []
        """The list of records currently managed by this manager."""
        self.store_entire_run: bool = store_entire_run
        """Flag for whether to store/copy environment info, log, output report, and
        all cached files in a run-specific folder."""
        self.dry: bool = dry
        """Flag for whether to suppress all file outputs (cached files, logs, registries etc.)"""
        self.dry_cache: bool = dry_cache
        """Flag for whether to suppress only writing cached files."""
        self.artifacts = []
        """The list of :code:`ArtifactRepresentation` instances for all artifacts stored in all
        record states."""
        self.reportables = []
        """The list of all reportables reported from all records."""

        self.stored = False
        """A flag keeping track of if the experiment :code:`store.json` has been updated with this run
        or not."""
        self.run_info = None
        """The metadata block associated with this manager from the :code:`ManagerStore`."""
        self.run_line = run_line
        """The CLI command used to run the current experiment."""
        self.reproduction_line = ""
        """The CLI command to replicate the experiment this manager is for."""
        self.interactive = False
        """Indicates if this manager was spawned with LIVE status in an interactive environment (e.g. a python
        terminal or Jupyter notebook.)"""

        # TODO: shouldn't we allow this to directly be passed into the constructor?
        self.overwrite_stages = []
        """The list of individual stages for which to ignore the cache."""
        self.overwrite = False
        """For live session managers where you don't wish to set overwrite on individual args, you can universely set the manager to overwrite by changing this flag to True."""  # TODO: (01/27/2022) take this into consideration in staging.

        self.error_thrown = False
        """A flag indicating whether an error was thrown by the experiment."""
        self.status = "incomplete" if status_override is None else status_override
        """The current status of the experiment: 'incomplete', 'complete', 'error', or 'LIVE' if run from a
        notebook, interactive terminal, or non-experiment script."""
        self.error = None
        """The exception class and error string, if one was thrown."""

        self.parallel_lock = parallel_lock
        """If this function is called from a multiprocessing context, use this lock to help
        prevent files being written to and read simultaneously."""
        self.parallel_mode = parallel_mode
        """A flag indicating whether this manager is in a "subproc" run or not. Do not set this manually."""

        self.lazy = lazy
        """If true, attempts to set all stage outputs as Lazy objects. Outputs that do not have.
        a cacher specified will be given a PickleCacher. Note that objects without a cacher tht do not
        handle pickle serialization correctly may cause errors."""
        self.ignore_lazy = ignore_lazy
        """Run the experiment disabling any lazy object caching/keeping everything in memory.
        This can save time when memory is less of an issue."""

        self.live_report_path_generated = False
        """A flag indicting if the default report's graphs/reportables folders exist, this helps prevent live displays from breaking if multiple display functions called."""
        self.live_report_paths = None

        self._load_config()

        if not self.dry and not self.dry_cache:
            if not os.path.exists(self.cache_path):
                os.makedirs(self.cache_path)
            if not os.path.exists(self.runs_path):
                os.makedirs(self.runs_path)

        if not self.dry:
            # TODO: (01/25/2022) maybe keep reports path check for when we're actually generating the report
            if not os.path.exists(self.reports_path):
                os.makedirs(self.reports_path)
            if not os.path.exists(self.logs_path):
                os.makedirs(self.logs_path)

        # handle populating alternate run line
        if self.status == "LIVE":
            if hasattr(sys, "ps1"):
                self.run_line = "(Interactive environment)"
                self.interactive = True
            else:
                self.run_line = " ".join(sys.argv)

        if not self.parallel_mode:
            # NOTE: by this point we don't technically have things like experiment args, but subsequent store calls will appropriately update it (see update_run in store.py) The reason this is important is so that the log init below has the correct run number etc.
            self.store()

        # start logging if from a live environment (otherwise experiment script handles this)
        if self.status == "LIVE" and not suppress_live_log and not self.parallel_mode:
            log_name = self.get_reference_name()
            log_path = os.path.join(self.logs_path, f"{log_name}.log")
            level = logging.DEBUG if live_log_debug else logging.INFO
            if self.dry:
                log_path = None
            utils.init_logging(log_path, level, False)

    def _load_config(self):
        """Populate any non-pre-existing path values with config values."""
        self.config = utils.get_configuration()
        if self.manager_cache_path is None:
            self.manager_cache_path = self.config["manager_cache_path"]
        if self.cache_path is None:
            self.cache_path = self.config["cache_path"]
        if self.runs_path is None:
            self.runs_path = self.config["runs_path"]
        if self.logs_path is None:
            self.logs_path = self.config["logs_path"]
        if self.notebooks_path is None:
            self.notebooks_path = self.config["notebooks_path"]
        if self.reports_path is None:
            self.reports_path = self.config["reports_path"]
        if self.report_css_path is None:
            self.report_css_path = self.config["report_css_path"]

    def store(self):
        """Update the ManagerStore with this manager's run metadata."""
        if self.dry:
            return
        if self.stored:
            # update
            store = ManagerStore(self.manager_cache_path)
            self.run_info = store.update_run(self)
            if self.store_entire_run:
                # update relevant run_info too with new_run
                if self.run_info is not None:
                    with open(
                        self.get_run_output_path("run_info.json"), "w"
                    ) as outfile:
                        json.dump(self.run_info, outfile, indent=4)
            return
        else:
            store = ManagerStore(self.manager_cache_path)
            self.run_info = store.add_run(self)
            self.stored = True
            if self.store_entire_run:
                self.write_run_env_output()

    def write_run_env_output(self):
        """Write all environment metadata to a run folder for a --store-full run."""
        self.pip_freeze = utils.get_pip_freeze()
        self.conda_env = utils.get_conda_env()
        self.os = utils.get_os()

        with open(
            self.get_run_output_path("requirements.txt"), "w", newline=""
        ) as outfile:
            outfile.write(self.pip_freeze)

        if self.conda_env != "":
            with open(
                self.get_run_output_path("environment.yml"), "w", newline=""
            ) as outfile:
                outfile.write(self.conda_env)

        with open(self.get_run_output_path("environment_meta.txt"), "w") as outfile:
            outfile.write(self.os)

        if self.run_info is not None:
            with open(self.get_run_output_path("run_info.json"), "w") as outfile:
                json.dump(self.run_info, outfile, indent=4)

    def get_all_argsets(self) -> List:
        """This is important to get argsets that aren't obtained through parameter files, e.g. in an interactive session."""
        master_list = []
        found_hashes = []
        for record in self.records:
            if record.args is not None:
                if record.args.hash not in found_hashes:
                    master_list.append(record.args)
                    found_hashes.append(record.args.hash)

        return master_list

    def get_grouped_reportables(self) -> Dict[str, List[Reportable]]:
        """Returns a dictionary of reportable groups, each group containing the list of reportables."""
        grouped_reportables = {}
        for reportable in self.reportables:
            if reportable.group is None:
                continue
            if reportable.group not in grouped_reportables:
                grouped_reportables[reportable.group] = []
            grouped_reportables[reportable.group].append(reportable)
        return grouped_reportables

    def get_ungrouped_reportables(self) -> List[Reportable]:
        """Returns the list of reportables that have no group."""
        non_grouped_reportables = []
        for reportable in self.reportables:
            if reportable.group is None:
                non_grouped_reportables.append(reportable)
        return non_grouped_reportables

    def get_path(
        self,
        obj_name: str,
        record: Record,
        output: bool = False,
        base_path: str = None,
        aggregate_records: List[Record] = None,
    ) -> str:
        """Get an appropriate full path/filename for a given object name and record.

        This is used by the cachers, it automatically handles generating a filename
        using appropriate experiment name prefixing etc.

        Args:
            obj_name (str): The name to associate with the object as the last part of the filename.
            record (Record): The record that this object is associated with. (Used to get experiment name, args hash
                and so on.)
            output (bool): Set this to true if the path needs to be based in a --store-full run folder.
            base_path (str): If a specific path override is needed, pass it in here. (Otherwise the
                manager's cache_path is used.)
            aggregate_records (List[Record]): If a list of records is passed (not none), prefix the
                path filename with the hash of arg hashes of all the passed records. This is used for
                paths for cached objects of aggregate stages.

        Returns:
            A string filepath that an object can be written to.
        """
        args_hash = "None"

        # compute args hash if necessary
        if record.args is not None:
            record.args.hash = utils.args_hash(
                record.args, self.manager_cache_path, self.dry
            )
            args_hash = record.args.hash
            object_path = f"{self.experiment_name}_{record.args.hash}_{self.current_stage_name}_{obj_name}"

            if self.store_entire_run:
                utils.args_hash(record.args, self.get_run_output_path(), self.dry)

        # set the hash to the hashed version of all args hashes from passed records if applicable
        if aggregate_records is not None:
            args_hash = utils.add_args_combo_hash(
                record, aggregate_records, self.manager_cache_path, not self.dry
            )  # TODO: uses not here, but doesn't up above?
            if self.store_entire_run:
                utils.add_args_combo_hash(
                    record, aggregate_records, self.get_run_output_path(), not self.dry
                )

        object_path = (
            f"{self._get_name()}_{args_hash}_{self.current_stage_name}_{obj_name}"
        )

        if output:
            return os.path.join(self.get_run_output_path(), object_path)
        elif base_path is not None:
            return os.path.join(base_path, object_path)
        else:
            return os.path.join(self.cache_path, object_path)

    def get_str_timestamp(self) -> str:
        """Convert the manager's run timestamp into a string representation."""
        return self.run_timestamp.strftime(utils.TIMESTAMP_FORMAT)

    def _get_name(self) -> str:
        if self.custom_name is None:
            return self.experiment_name
        return self.custom_name

    def get_reference_name(self) -> str:
        """Get the reference name of this run in the experiment registry.

        The format for this name is [experiment_name]_[run_number]_[timestamp]."""
        return f"{self.experiment_name}_{self.experiment_run_number}_{self.get_str_timestamp()}"

    def get_run_output_path(self, obj_name: str = None):
        """Get the path for a --store-full run folder for this manager. Similar to get_path, but
        with an always assumed output=True.

        Returns:
            A string filepath pointing to the runs folder with an optional object name attached at
            end. (Returns only the folder path if None is passed.)
        """
        output_path = os.path.join(self.runs_path, self.get_reference_name())
        return_path = output_path
        # have to be careful ordering of where we put stuff into output_path, because makedirs will
        # make _everything_ a dir, even filename a dir too, so only include if a '/' found in obj_name
        if obj_name is not None:
            return_path = os.path.join(return_path, obj_name)
            if "/" in obj_name:
                additional_sub_folders = obj_name[: obj_name.rindex("/")]
                output_path = os.path.join(output_path, additional_sub_folders)
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        return return_path

    def lock(self):
        """Necessary to avoid file collisions when run in parallel mode."""
        if self.parallel_lock is not None:
            self.parallel_lock.acquire()

    def unlock(self):
        """Necessary to avoid file collisions when run in parallel mode."""
        if self.parallel_lock is not None:
            self.parallel_lock.release()

    def generate_report(self):
        """Output report files to a run-specific report folder, the reports/_latest, and the store-full folder if applicable."""
        if self.dry:
            return
        logging.info("Generating report...")

        self.store()

        reporting.run_report(self, self.reports_path, "_latest", self.report_css_path)
        self.live_report_paths = reporting.run_report(
            self, self.reports_path, self.get_reference_name(), self.report_css_path
        )
        self.live_report_path_generated = True
        if self.store_entire_run:
            reporting.run_report(
                self, self.get_run_output_path(), "report", self.report_css_path
            )
        reporting.update_report_index(
            self.config["experiments_module_name"], self.reports_path
        )

    def display_info(self):
        """Returns an IPython HTML display object with the report info block. This is mostly only useful for making displays in a Juptyer notebook."""
        try:
            from IPython.display import HTML
        except ModuleNotFoundError:
            return "Unable to import IPython."
        self.store()

        return HTML("".join(reporting.render_report_info_block(self)))

    def _reportable_display_prep(self):
        from IPython.display import HTML  # TODO: better error handling for this?

        self.store()

        if not self.live_report_path_generated:
            self.live_report_paths = reporting.prepare_report_path(
                self.reports_path, self.get_reference_name()
            )
            self.live_report_path_generated = True
        return HTML

    def display_all_reportables(self):
        """Displays (via html) all produced reportables.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()
        folder_path, graphs_path, reportables_path = self.live_report_paths

        html = reporting.render_report_all_reportables(
            self,
            reportables_path,
            override_display_path=os.path.join("/", "files", reportables_path),
            notebook=True,
        )

        return HTML("".join(html))

    def display_record_reportables(self, record: Record):
        """Displays (via html) all reportables produced within the passed record.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()
        folder_path, graphs_path, reportables_path = self.live_report_paths

        html_lines = []
        for reportable in self.reportables:
            if reportable.record == record:
                html_lines.extend(
                    reporting.render_reportable(
                        reportable,
                        self,
                        reportables_path,
                        override_display_path=os.path.join(
                            "/", "files", reportables_path
                        ),
                        notebook=True,
                    )
                )

        return HTML("".join(html_lines))

    def display_group_reportables(self, group_name: str):
        """Displays (via html) all reportables in the passed group.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()
        folder_path, graphs_path, reportables_path = self.live_report_paths

        html_lines = []
        for reportable in self.reportables:
            if reportable.group == group_name:
                html_lines.extend(
                    reporting.render_reportable(
                        reportable,
                        self,
                        reportables_path,
                        override_display_path=os.path.join(
                            "/", "files", reportables_path
                        ),
                        notebook=True,
                    )
                )

        return HTML("".join(html_lines))

    def display_stage_reportables(self, stage_name: str):
        """Displays (via html) all reportables produced by the given stage.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()
        folder_path, graphs_path, reportables_path = self.live_report_paths

        html_lines = []
        for reportable in self.reportables:
            if reportable.stage == stage_name:
                html_lines.extend(
                    reporting.render_reportable(
                        reportable,
                        self,
                        reportables_path,
                        override_display_path=os.path.join(
                            "/", "files", reportables_path
                        ),
                        notebook=True,
                    )
                )

        return HTML("".join(html_lines))

    def display_reportable(self, reportable: Reportable):
        """Displays (via html) the rendered passed reportable.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()
        folder_path, graphs_path, reportables_path = self.live_report_paths

        html = reporting.render_reportable(
            reportable,
            self,
            reportables_path,
            override_display_path=os.path.join("/", "files", reportables_path),
            notebook=True,
        )

        return HTML("".join(html))

    def display_stage_graph(self):
        """Displays (via html) the graphviz SVG map of the records and the stages they were run through.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()

        html = reporting.render_graph(reporting.map_full_svg(self))
        return HTML("".join(html))

    def get_reportable_groups(self) -> List[str]:
        groups = []
        for reportable in self.reportables:
            if reportable.group not in groups:
                groups.append(reportable.group)
        return groups



"""This is the 'main' runnable function and CLI, which handles setting up logging,
folders, reports, and running the passed experiment.

This file contains a :code:`__name__ == "__main__"` and can be run directly.
"""

import argparse
import datetime
import importlib
import glob
import logging
import multiprocessing as mp
import os
import re
import shutil
import subprocess
import sys
import traceback

from typing import List

from curifactory.manager import ArtifactManager
from curifactory import utils, docker

CONFIGURATION_FILE = "curifactory_config.json"


def run_experiment(  # noqa: C901 -- TODO: this does need to be broken up at some point
    experiment_name,
    parameters_list,
    overwrite_override=None,
    cache_dir_override=None,
    mngr: ArtifactManager = None,
    log: bool = False,
    log_debug: bool = False,
    dry: bool = False,
    dry_cache: bool = False,
    store_entire_run: bool = False,
    log_errors: bool = False,
    custom_name: str = None,
    build_docker: bool = False,
    build_notebook: bool = False,
    run_string: str = None,
    stage_overwrites: List[str] = None,
    args_names: List[str] = None,
    args_indices: List[str] = None,
    global_args_indices: List[str] = None,
    parallel: int = None,
    parallel_mode: bool = False,
    parallel_lock: mp.Lock = None,
    parallel_queue: mp.Queue = None,
    run_num_override: int = None,
    run_ts_override: datetime.datetime = None,
    lazy: bool = False,
    ignore_lazy: bool = False,
    notes: str = None,
):
    """The experiment entrypoint function. This executes the given experiment
    with the given parameters.

    Args:
        experiment_name (str): The name of the experiment script (without the :code:`.py`).
        parameters_list (List[str]): A list of names of parameter files (without the :code:`.py`).
        overwrite_override (bool): Whether to force overwrite on all cache data.
        cache_dir_override (str): Specify a non-default cache location. This would be used if
            running with the cache from a previous --store-full run.
        mngr (ArtifactManager): An artifact manager to use for the experiment. One will be automatically
            created if none is passed.
        log (bool): Whether to write a log file or not.
        log_debug (bool): Whether to include DEBUG level messages in the log.
        dry (bool): Setting dry to true will suppress saving any files (including logs), and will
            not update parameter stores. (It should have no effect on any files.)
        dry_cache (bool): Setting this to true only suppresses saving cache files. This is recommended
            if you're running with a cache_dir_override for some previous --store-full run, so you
            don't accidentally overwrite or add new data to the --store-full directory.
        store_entire_run (bool): Store environment info, log, output report, and all cached files in a
            run-specific folder (:code:`data/runs` by default)
        log_errors (bool): Whether to include error messages in the log output.
        custom_name (str): Instead of using the experiment name to group cached data, use this name instead.
        build_docker (bool): If true, build a docker image with all of the run cache afterwards.
        build_notebook (bool): If true, add a notebook with run info and default cells to reproduce
            after run execution.
        run_string (str): An automatically populated string representing the CLI command for the run,
            do not change this.
        stage_overwrites (List[str]): A list of string stage names that you should overwrite, this is
            useful if there are specific stages you sometimes want to recompute but the remainder of
            the data can remain cached.
        args_names (List[str]): A list of argument names to run. If this is specified only arguments
            returned from the passed parameters, with these names, will be passed to the experiment.
        args_indices (List[str]): A list of argument indices to run. If this is specified, only the
            arguments returned from the passed parameter files, indexed by the ranges specified, will
            be passed to the experiment. Note that you can specify ranges delineated with '-', e.g. '3-7'.
        global_args_indices (List[str]): A list of argument indices to run, indexing the entire set
            of parameters passed to the experiment instead of each individual paarameters file. This
            can be used to help more intelligently parallelize runs. Formatting follows the same rules
            as args_indices.
        parallel (int): How many subprocesses to split this run into. If specified, the experiment will
            be run that many times with divided up global_args_indices in order to generated cached
            data for all parameters, and then re-run a final time with all cached data combined. Note
            then that any speedup from this is based on how well and how many steps are cached.
        parallel_mode (bool): This is handled by the parallel parameter, informing a particular subproc
            run that it is being executed from a parallel run.
        parallel_lock (multiprocessing.Lock): If this function is called from a multiprocessing context, use
            this lock to help prevent files being written to and read simultaneously.
        parallel_queue (multiprocessing.Queue): If this function is called from a multiprocessing context,
            use this queue for communicating success/errors back to the main process.
        run_num_override (int): Handled in parallel mode, since parallel process experiments do not get
            run number (since manager.store is not called) the log gets stored in an incorrectly named
            file.
        run_ts_override (datetime.datetime): Handled in parallel mode, since parallel process experiments
            do not get the same timestamp if started a few seconds later, the log gets stored in an
            incorrectly named file.
        lazy (bool): If true, attempts to set all stage outputs as Lazy objects. Outputs that do not have
            a cacher specified will be given a PickleCacher. Note that objects without a cacher that do not
            handle pickle serialization correctly may cause errors.
        ignore_lazy (bool): Run the experiment disabling any lazy object caching/keeping everything in memory.
            This can save time when memory is less of an issue.
        notes (str): A git-log-like message to store in the run info for the current run. If this is an
            empty string, query the user for an input string.

    Returns:
        Whatever is returned from the experiment :code:`run()`.

    Example:
        .. code-block:: python

            from curifactory import experiment, ArtifactManager
            mngr = ArtifactManager()
            experiment.run_experiment('exp_name', ['params1', 'params2'], mngr=mngr, dry=True)
    """

    if run_string is None:
        run_string = f"experiment {experiment_name}"
        for param_file in parameters_list:
            run_string += f" -p {param_file}"
        if args_names is not None:
            for arg_name in args_names:
                run_string += f" --names {arg_name}"
        if args_indices is not None:
            for arg_index in args_indices:
                run_string += f" --indices {arg_index}"
        if stage_overwrites is not None:
            for stage_overwrite in stage_overwrites:
                run_string += f" --overwrite-stage {stage_overwrite}"
        if cache_dir_override is not None:
            run_string += f" --cache {cache_dir_override}"
        if overwrite_override:
            run_string += " --overwrite"
        if parallel is not None:
            run_string += f" --parallel {parallel}"
        if parallel_mode:
            run_string += " --parallel-mode"
        if global_args_indices is not None:
            for index_range in global_args_indices:
                run_string += f" --global-indices {index_range}"
        if store_entire_run:
            run_string += " --store-full"
        if dry:
            run_string += " --dry"
        if dry_cache:
            run_string += " --dry-cache"
        if custom_name is not None:
            run_string += f" --name {custom_name}"
        if lazy:
            run_string += " --lazy"
        if ignore_lazy:
            run_string += " --ignore-lazy"

        # TODO: remainder of flags

    # NOTE: this allows a project root /params /experiments to work, but unfortunately doesn't automagically work if you have nested folders (you need to put __init__.py in each dir)
    sys.path.append(os.getcwd())

    # distributed run check, automatically set parallel mode if we're not rank 0
    # this was added because of issues with pytorch distributed compute
    if "LOCAL_RANK" in os.environ:
        if os.getenv("LOCAL_RANK") != "0":
            parallel_mode = True
        elif "NODE_RANK" in os.environ and os.getenv("NODE_RANK") != "0":
            parallel_mode = True

    # get the experiment run notes if requested
    if notes == "":
        notes_file = "temp_run_notes_content.txt"
        # create notes template
        with open(notes_file, "w") as outfile:
            outfile.write(
                "\n\n"
                + '# Write a "git-commit-like" message to attach as notes to the experiment run. The first line will be treated as the shortform notes line.'
                + "\n# All lines starting with a '#' will be ignored."
            )

        # try to find an appropriate editor
        os.system(f"{utils.get_editor()} {notes_file}")

        # get the content
        with open(notes_file, "r") as infile:
            lines = [line for line in infile.readlines() if not line.startswith("#")]
        notes = "".join(lines).strip("\n")

        os.remove(notes_file)

    # force full store if building a docker container
    if build_docker:
        store_entire_run = True

    if mngr is None:
        mngr = ArtifactManager(
            experiment_name,
            store_entire_run=store_entire_run,
            dry=dry,
            dry_cache=dry_cache,
            custom_name=custom_name,
            run_line=run_string,
            parallel_lock=parallel_lock,
            parallel_mode=parallel_mode,
            lazy=lazy,
            ignore_lazy=ignore_lazy,
            status_override=None,
            notes=notes,
        )
        # mngr.experiment_name = experiment_name
        mngr.experiment_args_file_list = parameters_list
        if run_num_override is not None:
            mngr.experiment_run_number = run_num_override
        if run_ts_override is not None:
            mngr.run_timestamp = run_ts_override

        if stage_overwrites is not None:
            # TODO: only apply to subprocs if parallel?
            mngr.overwrite_stages = stage_overwrites

    # check for cache path override
    if cache_dir_override is not None:
        mngr.cache_path = cache_dir_override

    # resolve args indices and any specified ranges
    args_indices_resolved = []
    if args_indices is not None:
        for entry in args_indices:
            if "-" in entry:
                start = int(entry[: entry.index("-")])
                end = int(entry[entry.index("-") + 1 :])
                args_indices_resolved.extend(list(range(start, end)))
            else:
                args_indices_resolved.append(int(entry))

    # resolve global args indices and any specified ranges
    global_args_indices_resolved = []
    if global_args_indices is not None:
        for entry in global_args_indices:
            if "-" in entry:
                start = int(entry[: entry.index("-")])
                end = int(entry[entry.index("-") + 1 :])
                global_args_indices_resolved.extend(list(range(start, end)))
            else:
                global_args_indices_resolved.append(int(entry))

    # handling logging. NOTE: we do have to keep this here because of parallel considerations
    if log:
        log_name = mngr.get_reference_name()
        if parallel_mode:
            if global_args_indices is not None:
                log_name += f"_{str(global_args_indices[0])}"
            else:
                log_name += f"_p{str(os.getpid())}"
        log_path = os.path.join(mngr.logs_path, f"{log_name}.log")

        if dry:
            log_path = None

        level = logging.INFO
        if log_debug:
            level = logging.DEBUG
        utils.init_logging(log_path, level, log_errors, include_process=parallel_mode)

    logging.info("Running experiment %s" % experiment_name)

    # insert an implied parameter file that is the same as the experiment file, if no parameters were explicitly listed
    if parameters_list is None or len(parameters_list) == 0:
        logging.info(
            "No parameter files listed, assuming an implied experiment file get_params(), inserting '%s'"
            % experiment_name
        )
        parameters_list = [experiment_name]

    # load params files
    argsets = []
    for params in parameters_list:
        mngr.experiment_args[params] = []
        # TODO: if we get a ModuleNotFoundError, suggest ensuring __init__.py as appropriate and that module paths don't have '/'
        param_module_string = f"{mngr.config['params_module_name']}.{params}"
        experiment_module_string = f"{mngr.config['experiments_module_name']}.{params}"
        try:
            logging.debug("Trying to load params module '%s'" % param_module_string)
            param_module = importlib.import_module(param_module_string)
        except ModuleNotFoundError:
            # TODO:
            try:
                logging.debug(
                    "Module not found, trying '%s'" % experiment_module_string
                )
                param_module = importlib.import_module(experiment_module_string)
            except ModuleNotFoundError as e:
                logging.error(
                    "Parameter file '%s' could not be found in either experiments or parameters directory. Ensure curifactory_config.json is correct and if module subpaths are used, try including an __init__.py in each folder."
                    % params
                )
                raise e

        param_argsets = param_module.get_params()
        argsets_to_add = []

        # NOTE: we don't want to set override in parent proc on parallel runs.
        if overwrite_override and parallel is None:
            for argset in param_argsets:
                argset.overwrite = True

        # TODO: if param_argsets is not of type List, throw exception and advise
        # compute the hash of every argset and store the params
        for index, argset in enumerate(param_argsets):

            # if specific names requested, just grab those
            if args_names is not None:
                if argset.name not in args_names:
                    continue

            # if specific indices requested, just grab those
            if len(args_indices_resolved) > 0:
                if index not in args_indices_resolved:
                    continue

            argset.hash = utils.args_hash(
                argset, mngr.manager_cache_path, (not dry and not parallel_mode)
            )
            mngr.experiment_args[params].append((argset.name, argset.hash))
            # TODO: (01/24/2022) I have no idea what the point of this args_hash is...
            # it's not storing anything, and args_hash has no side-effects, so unclear
            # on why this matters.
            if store_entire_run:
                utils.args_hash(
                    argset, None, False
                )  # don't try to store because get_run_output_path does not exist yet
            argsets_to_add.append(argset)

        argsets.extend(argsets_to_add)

    if len(global_args_indices_resolved) > 0:
        argsets = [argsets[i] for i in global_args_indices_resolved]

    if not parallel_mode:
        mngr.store()

    if not parallel_mode:
        # this has to be done AFTER mngr.store, otherwise the reference name does not have the correct run number.
        logging.info(
            "Experiment run reference name is '%s'" % mngr.get_reference_name()
        )
    logging.info("Run command is '%s'" % run_string)

    # store params registry in run folder
    # TODO: is this already being done in manager get_path?
    if store_entire_run:
        for argset in argsets:
            utils.args_hash(
                argset, mngr.get_run_output_path(), (not dry and not parallel_mode)
            )

    # note that nothing is being cached or stored
    if dry:
        logging.info(
            "NOTE - running in dry mode. This log will not be saved, and no new"
            + " files will be cached. (Existing caches will still be read.)"
        )
    if store_entire_run:
        logging.info(
            "NOTE - running in full store mode. A copy of all cached objects and environment information will be stored in %s",
            mngr.get_run_output_path(),
        )

    # handle parallelism
    if parallel is not None:
        logging.info("Running experiment in parallel on %s processes..." % parallel)
        # divide up args list
        countper = int(len(argsets) / parallel)
        counts = [countper] * parallel
        remainder = len(argsets) % parallel
        index = 0
        while remainder > 0:
            counts[index] += 1
            remainder -= 1
            index += 1
            if index >= len(argsets):
                index = 0

        # get the formatted string range for each thread
        global_index_ranges = []
        start = 0
        for count in counts:
            end = start + count
            if count == 0:
                logging.warning(
                    "More processes specified than argsets, will run with fewer processes."
                )
            else:
                global_index_ranges.append([f"{start}-{end}"])
            start = end
        logging.info("Parallel arg splits: %s", str(global_index_ranges))

        processes = []
        multiprocessing_lock = mp.Lock()
        multiprocessing_queue = mp.Queue()
        for index_range in global_index_ranges:
            p = mp.Process(
                target=run_experiment,
                args=(
                    experiment_name,
                    parameters_list,
                    overwrite_override,
                    cache_dir_override,
                    None,  # mngr
                    log,
                    log_debug,
                    dry,
                    dry_cache,
                    False,  # store_entire_run
                    log_errors,
                    custom_name,
                    False,  # build_docker
                    False,  # build_notebook
                    None,  # run_string
                    stage_overwrites,
                    args_names,  # args_names
                    args_indices,  # args_indices
                    index_range,
                    None,  # parallel
                    True,  # parallel_mode
                    multiprocessing_lock,
                    multiprocessing_queue,
                    mngr.experiment_run_number,
                    mngr.run_timestamp,
                    lazy,
                    ignore_lazy,
                ),
            )
            logging.info(
                "Starting new experiment process for parameter range %s...", index_range
            )
            p.start()
            processes.append(p)

        for i, p in enumerate(processes):
            # TODO - get success/fail status
            p.join()
            run_status = multiprocessing_queue.get()
            if run_status[1] == "error":
                mngr.status = "error"
                mngr.error = run_status[2]
                mngr.error_thrown = True
                logging.info(
                    "Joined process for range '%s' with status code '%s' - %s"
                    % (global_index_ranges[i], run_status[1], run_status[2])
                )
            else:
                logging.info(
                    "Joined process for range '%s' with status code '%s'"
                    % (global_index_ranges[i], run_status[1])
                )

        logging.info(
            "All experiment processes have re-joined, continuing final combined run..."
        )

    if parallel_mode:
        logging.info(
            "Running experiment process for parameter range %s"
            % str(global_args_indices)
        )

    # run the experiment
    error_thrown = False
    experiment_module = importlib.import_module(
        f"{mngr.config['experiments_module_name']}.{experiment_name}"
    )
    try:
        results = experiment_module.run(argsets, mngr)

        # don't change status if we logged an error from a parallel process
        if not mngr.error_thrown:
            mngr.status = "complete"
    except Exception as e:
        results = None
        error_thrown = True
        logging.error(e)
        logging.error(traceback.format_exc())

        # don't change status if we logged an error from a parallel process
        if not mngr.error_thrown:
            mngr.status = "error"
            mngr.error = f"{str(e.__class__.__name__)} - {str(e)}"

    # send the results from a parallel process to the main process
    if parallel_mode and parallel_queue is not None:
        parallel_queue.put((global_args_indices[0], mngr.status, mngr.error))

    if not parallel_mode:
        mngr.store()

    if log and store_entire_run and not dry:
        # copy the logfile(s) over (when running in parallel, grab the subproc logs too)
        log_name = mngr.get_reference_name()
        log_path = f"logs/{log_name}*.log"
        for file in glob.glob(log_path):
            shutil.copy(file, mngr.get_run_output_path())
        # shutil.copyfile(
        #     log_path, os.path.join(mngr.get_run_output_path(), f"{log_name}.log")
        # )

        # TODO details file for any meta information (if it errored or not)

    if build_notebook and not dry and not parallel_mode:
        try:
            os.mkdir("notebooks/experiments")
        # TODO: Handle exceptions better
        except:  # noqa: E722
            pass
        write_experiment_notebook(
            experiment_name,
            parameters_list,
            argsets,
            mngr,
            path="notebooks/experiments/" + mngr.get_reference_name(),
            use_global_cache=None,
            errored=error_thrown,
        )

    if build_docker and not dry and not parallel_mode:
        if error_thrown:
            logging.warning("The experiment threw an error - docker build skipped.")
        else:
            write_experiment_notebook(
                experiment_name,
                parameters_list,
                argsets,
                mngr,
                mngr.get_run_output_path("run_notebook"),
                use_global_cache=True,
                directory_change_back_depth=1,
                suppress_global_warning=True,
            )
            docker.build_docker(
                experiment_name,
                mngr.get_run_output_path()[:-1],
                f"{mngr.experiment_run_number}_{mngr.run_timestamp.strftime('%Y-%m-%d')}",
            )

    if not dry and not parallel_mode:
        mngr.generate_report()

    return results, mngr


def write_experiment_notebook(
    experiment_name,
    parameters_list,
    argsets,
    manager,
    path,
    directory_change_back_depth=2,
    use_global_cache=None,
    errored=False,
    suppress_global_warning=False,
):
    """Creates a jupyter notebook prepopulated with experiment info and cells to re-run
    the experiment and discover all associated data stored in record states. This function
    is run by the :code:`run_experiment()` function.

    Args:
        experiment_name (str): The name of the run experiment
        parameters_list (List[str]): List of parameter file names
        argsets (List[Args]): List of all used :code:`Args` from parameter files.
        manager (ArtifactManager): :code:`ArtifactManager` used in the experiment.
        path (str): The path to the directory to store the notebook in.
        directory_change_back_depth (int): How many directories up the notebook needs
            to be in the project root (so imports and cache paths are all correct.)
        use_global_cache (bool): Whether we're using the normal experiment cache or
            a separate specific cache folder (mostly just used to display a warning
            in the notebook.)
        errored (bool): Whether the experiment errored or nat while running, will display
            a warning in the notebook.
        suppress_global_warning (bool): Don't show a warning if :code:`use_global_cache`
            is true.
    """
    # TODO - move elsewhere?
    logging.info("Creating experiment notebook...")

    if use_global_cache is None:
        use_global_cache = not manager.store_entire_run

    output_lines = [
        "# %%",
        "'''",
        f"# {manager.experiment_name} - {manager.experiment_run_number}",
        f"\nExperiment name: **{manager.experiment_name}**  ",
        f"Experiment run number: **{manager.experiment_run_number}**  ",
        f"Run timestamp: **{manager.run_timestamp.strftime('%m/%d/%Y %H:%M:%S')}**  ",
        f"Reference: **{manager.get_reference_name()}**  ",
        f"Git commit: {manager.git_commit_hash}  ",
        f"Params files: {str(manager.experiment_args_file_list)}",
        "\n**Parameters**:",
    ]

    # output the list of parameters used and assoc hashes
    for key in manager.experiment_args:
        output_lines.append(f"* {key}")
        for name, hash in manager.experiment_args[key]:
            output_lines.append(f"\t* {name} - {hash}")

    output_lines.extend(
        [
            "---",
            "'''",
            "",
            # "# %%",
            # "'''",
        ]
    )

    # NOTE: for large argsets it makes this cell take up way too much room
    # # output detailed args for each paramset
    # for key in manager.experiment_args:
    #     for name, hash in manager.experiment_args[key]:
    #         output_lines.append(f"\n**{name} - {hash}**")
    #         output_lines.append("```")
    #         for argset in argsets:
    #             if argset.hash == hash:
    #                 argset_data = asdict(argset)
    #                 del argset_data["name"]
    #                 del argset_data["overwrite"]
    #                 del argset_data["hash"]
    #                 output_lines.append(
    #                     json.dumps(argset_data, indent=4, default=lambda x: str(x))
    #                 )
    #                 break
    #         output_lines.append("```")

    # output_lines.extend(["---", "'''", ""])

    # pathing for whether full run or not
    directory_change_back = "/".join([".."] * directory_change_back_depth)
    cache_dir_arg = f"manager_cache_path='{manager.get_run_output_path()}', cache_path='{manager.get_run_output_path()}', "
    # warn if data is potentially wrong
    if use_global_cache:
        cache_dir_arg = ""
        if not suppress_global_warning:
            output_lines.extend(
                [
                    "# %%",
                    "'''",
                    "<span style='color: orange;'><b>WARNING - </b></span>Experiment was not run with a `--store-full` flag, and so is simply using the project-wide cache rather than a specific experiment run cache. Any recent experiment runs since this notebook was created may have altered cached data."
                    "'''",
                    "",
                ]
            )

    if errored:
        output_lines.extend(
            [
                "# %%",
                "'''",
                "<span style='color: red;'><b>WARNING - </b></span>This experiment run did not complete due to an exception."
                "'''",
                "",
            ]
        )

    # imports and logger lines
    output_lines.extend(
        [
            "# %%",
            f"%cd {directory_change_back}",
            "",
            "# %%",
            "import logging",
            "from curifactory import ArtifactManager, experiment",
            "",
            "# %%",
            "logger = logging.getLogger()",
            "logger.setLevel(logging.INFO)",
            "",
            "# %%",
            f'manager = ArtifactManager("{experiment_name}", {cache_dir_arg} dry=True)',
            f'experiment.run_experiment("{experiment_name}", {str(parameters_list)}, dry=True, mngr=manager)',
            "",
            "# %%",
        ]
    )

    for i, param in enumerate(manager.records):
        output_lines.extend(
            [
                f"records{i} = manager.records[{i}]",
                f"state{i} = manager.records[{i}].state",
            ]
        )

        if manager.records[i].args is not None:
            output_lines.extend(
                [
                    f'print("state{i} - (" + records{i}.args.name + ") stages: " + str(records{i}.stages))',
                    f'print("keys: " + str(state{i}.keys()) + "\\n")',
                ]
            )
        else:
            output_lines.extend(
                [
                    f'print("state{i} - ((aggregate record)) stages: " + str(records{i}.stages))',
                    f'print("keys: " + str(state{i}.keys()) + "\\n")',
                ]
            )
        output_lines.append("")

    script_path = path + ".py"
    notebook_path = path + ".ipynb"

    output_lines = [line + "\n" for line in output_lines]

    with open(script_path, "w") as outfile:
        outfile.writelines(output_lines)

    # run ipynb-py-convert
    logging.info("Converting...")
    # utils.get_command_output(["ipynb-py-convert", script_path, notebook_path])

    cmd_array = ["ipynb-py-convert", script_path, notebook_path]
    print(*cmd_array)
    with subprocess.Popen(
        cmd_array, stderr=subprocess.PIPE, stdout=subprocess.PIPE, bufsize=1, text=True
    ) as p:
        for line in p.stdout:
            print(line, end="")  # process line here
        for line in p.stderr:
            print(line, end="")  # process line here
    logging.info("Output experiment notebook at %s", notebook_path)
    os.remove(script_path)


def regex_lister(path, regex):
    """Used by both list_experiments and list_params. This scans every file in the passed
    folder for the requested regex, and tries to import the files that have a match."""

    names = []

    if not os.path.exists(path):
        print(
            f"\t[WARNING - path: '{path}' does not exist. Double check curifactory_config.json]"
        )
        return []

    for filename in os.scandir(path):
        # iterate every python file
        if filename.path.endswith(".py"):
            with open(filename.path, "r") as infile:
                lines = infile.readlines()
                # check for a run() method
                results = [re.findall(regex, line) for line in lines]

                # remove empty arrays from findall search
                clean_results = []
                for result in results:
                    if result != []:
                        clean_results.extend(result)

                # did we find a run()?
                if len(clean_results) > 0:
                    non_pyextension_name = filename.name[:-3]

                    # see if it's valid
                    try:
                        importlib.import_module(f"{path}.{non_pyextension_name}")
                        comment = utils.get_py_opening_comment(lines)
                        if comment != "":
                            names.append(
                                non_pyextension_name
                                + " - "
                                + utils.get_py_opening_comment(lines)
                            )
                        else:
                            names.append(non_pyextension_name)
                    except Exception as e:
                        names.append(non_pyextension_name + " [ERROR - " + str(e) + "]")

    return names


def list_experiments():
    """Print out all valid experiments that have a :code:`def run()` function, including
    any top-of-file docstrings associated with each."""
    config = utils.get_configuration()

    experiment_names = regex_lister(
        config["experiments_module_name"], r"^def run\(.*\)"
    )

    return experiment_names


def list_params():
    """Print out all valid parameter files that have a :code:`def get_params()`
    function, including any top-of-file docstrings associated with each."""
    config = utils.get_configuration()

    param_names = []

    param_names.extend(
        regex_lister(config["params_module_name"], r"^def get_params\(.*\)")
    )
    param_names.extend(
        regex_lister(config["experiments_module_name"], r"^def get_params\(.*\)")
    )
    return param_names


def main():
    """'Main' command line entrypoint, parses command line flags and makes the
    appropriate :code:`run_experiment()` call."""
    parser = argparse.ArgumentParser(
        epilog="Example:\n\texperiment my_experiment -p some_params1 -p some_params2"
    )
    parser.add_argument("experiment_name")
    parser.add_argument(
        "-p",
        dest="parameters_name",
        # required=True,
        action="append",
        help="The name of a parameters python file. Does not need to include path or .py. You can specify multiple -p arguments to run all parameters",
    )
    parser.add_argument(
        "--names",
        dest="args_name",
        action="append",
        help="The name of a specific argset within one of the specified parameters files. Using this will run only the specified argset(s).",
    )
    parser.add_argument(
        "--indices",
        dest="args_index",
        action="append",
        help="A single or range of indices of argsets to use within the specified parameters files. Note that specifying this will run those indices from all specified parameters files. Specify ranges like '1-5'.",
    )
    parser.add_argument(
        "--global-indices",
        dest="global_indices",
        action="append",
        help="A single or range of indices of argsets to use out of the total list of argsets passed to the experiment.",
    )
    parser.add_argument(
        "--parallel",
        dest="parallel",
        default=None,
        help="Divide the parameters among n subprocesses in order to cache intermediate results in parallel, and then re-run in original process with cached data.",
    )
    parser.add_argument(
        "--parallel-mode",
        dest="parallel_mode",
        action="store_true",
        help="Runs an experiment and caches intermediate values but does not report or touch experiment store and parameter registry. Note that using this flag manually is not equivalent to running an experiment with the --parallel flag, as file locks don't get used when calling this manually, potentially creating race conditions.",
    )
    parser.add_argument(
        "--no-log",
        dest="no_log",
        action="store_true",
        help="Specify this flag to not store the log.",
    )
    parser.add_argument(
        "--overwrite",
        dest="overwrite_override",
        action="store_true",
        help="Overwrite current cached objects",
    )
    parser.add_argument(
        "--overwrite-stage",
        dest="overwrite_stages",
        action="append",
        help="Names of specific stages to overwrite",
    )
    parser.add_argument(
        "-s",
        "--store-full",
        dest="store_full",
        action="store_true",
        help="Store a copy of environment information and every cached object in a folder named with the experiment reference.",
    )
    parser.add_argument(
        "-c",
        "--cache",
        dest="cache_dir",
        help="Specify a different directory to use as the cache.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Log at debug level.",
    )
    parser.add_argument(
        "--dry",
        dest="dry",
        action="store_true",
        help="Do a dry run: suppresses writing any cached objects and modifying stores.",
    )
    parser.add_argument(
        "--dry-cache",
        dest="dry_cache",
        action="store_true",
        help="Do a dry cache run: this still modifies stores and runs reports but does not write anything into the cache. This is recommended when running an experiment with a cache directory from --store-full.",
    )
    parser.add_argument(
        "--log-errors",
        dest="log_errors",
        action="store_true",
        help="Include errors and stack traces in output logs. NOTE: this redirects stderr - output from libraries like tqdm will be logged as well.",
    )
    parser.add_argument(
        "-n",
        "--name",
        dest="name",
        default=None,
        help="Specify a custom name to use for caching rather than the experiment name. This can be useful if multiple similar experiments can use the same cached objects. Note that this does not change the reference name.",
    )
    parser.add_argument(
        "--docker",
        dest="docker",
        action="store_true",
        help="Generates a docker image with the used cached files and a current copy of the codebase.",
    )
    parser.add_argument(
        "--notebook",
        dest="notebook",
        action="store_true",
        help="Generates a jupyter notebook to explore the results.",
    )
    parser.add_argument(
        "--lazy",
        dest="lazy",
        action="store_true",
        help="Attempts to lazy-cache all outputs from all stages. WARNING: if cachers are not specified throughout the stages, PickleCachers will be inserted. As pickle serialization does not work on some complex objects, this may fail. Specify valid cachers for all stage outputs to mitigate this.",
    )
    parser.add_argument(
        "--ignore-lazy",
        dest="ignore_lazy",
        action="store_true",
        help="Run the experiment without any lazy object caching, this can save time when memory is less of an issue.",
    )
    parser.add_argument(
        "--notes",
        nargs="?",  # this is how we allow it to be specified either as a flag or with a value
        dest="notes",
        const="",
        default=False,
        help="Associate some notes with this run. You can either directly specify a string to this flag, or leave it blank to open a notes file in your editor.",
    )
    parser.add_argument(
        "--port",
        dest="port",
        default=8080,
        help="Only used for 'experiment reports', specifies which port to run the simple server on.",
    )
    parser.add_argument(
        "--host",
        dest="host",
        default="127.0.0.1",
        help="Only used for 'experiment reports', specifies which hostname to run the simple server on.",
    )

    # fix any missing quotes in run line
    command_parts = sys.argv[1:]

    fixed_parts = []
    for index, part in enumerate(command_parts):
        if " " in part and not part.startswith('"') and not part.endswith('"'):
            part = f'"{part}"'

            # for now we don't include the notes param, because the notes are already included in
            # the report render.
            if index > 0 and command_parts[index - 1] == "--notes":
                continue
        if part == "--notes":
            continue
        fixed_parts.append(part)

    run_string = "experiment " + " ".join(fixed_parts)

    args = parser.parse_args()
    params_list = args.parameters_name

    if args.experiment_name == "ls":

        sys.path.append(os.getcwd())
        print("EXPERIMENTS:")
        for experiment in list_experiments():
            print("\t" + experiment)
        print("\nPARAMS:")
        for param in list_params():
            print("\t" + param)
        exit()

    elif args.experiment_name == "reports":
        os.chdir("reports")
        utils.run_command(
            ["python", "-m", "http.server", str(args.port), "--bind", args.host]
        )
        os.chdir("..")
        exit()

    # todo: verify names exist
    # todo: ensure folders exist (logs)
    # todo: add cache overwrite option

    log = True
    if args.no_log:
        log = False

    parallel = None
    if args.parallel is not None:
        parallel = int(args.parallel)

    if args.notes != "" and not args.notes:
        args.notes = None

    run_experiment(
        args.experiment_name,
        params_list,
        args.overwrite_override,
        args.cache_dir,
        log=log,
        log_debug=args.verbose,
        dry=args.dry,
        dry_cache=args.dry_cache,
        store_entire_run=args.store_full,
        log_errors=args.log_errors,
        custom_name=args.name,
        build_docker=args.docker,
        build_notebook=args.notebook,
        run_string=run_string,
        stage_overwrites=args.overwrite_stages,
        args_names=args.args_name,
        args_indices=args.args_index,
        global_args_indices=args.global_indices,
        parallel=parallel,
        parallel_mode=args.parallel_mode,
        lazy=args.lazy,
        ignore_lazy=args.ignore_lazy,
        notes=args.notes,
    )


if __name__ == "__main__":
    main()



import pytest
from curifactory import (
    stage,
    aggregate,
    Record,
    InputSignatureError,
    OutputSignatureError,
    EmptyCachersError,
    CachersMismatchError,
)
from curifactory.caching import PickleCacher, Lazy


# --------------------------
# @stage tests
# --------------------------


def test_stores_output_in_record(configured_test_manager):
    """Returned output from stage function should exist in record state."""

    @stage([], ["tester"])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    output_stage(record)
    assert record.state["tester"] == "hello world"


def test_stage_outputs_none(configured_test_manager):
    """Stage outputs set to None should be equivalent to [] and should not crash."""

    @stage([], outputs=None)
    def output_stage(record):
        x = 5
        del x

    record = Record(configured_test_manager, None)
    output_stage(record)


def test_stage_inputs_none(configured_test_manager):
    """Stage inputs set to None should be equivalent to [] and should not crash."""

    @stage(inputs=None, outputs=["tester"])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    output_stage(record)
    assert record.state["tester"] == "hello world"


def test_stores_multiple_outputs_in_record(configured_test_manager):
    """Multiple returned outputs from stage function should exist in record state."""

    @stage([], ["test1", "test2"])
    def output_stage(record):
        return "hello world", 13

    record = Record(configured_test_manager, None)
    output_stage(record)
    assert record.state["test1"] == "hello world"
    assert record.state["test2"] == 13


def test_returns_less_than_expected_errors(configured_test_manager):
    """A function that doesn't return the same number of objects as specified in the stage outputs should throw an OutputSignatureError."""

    @stage([], ["test1", "test2"])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        output_stage(record)


def test_empty_cachers_array_errors(configured_test_manager):
    """Don't let the user write [] for cachers list, as it always shortcircuits."""
    # NOTE: technically this could probably just be a special case handled by staging.py, but I think specifying [] for cachers is kind of unclear anyway.
    @stage([], [], [])
    def stage_that_does_nothing(record):
        i_am_very_important = 5 + 3
        del i_am_very_important

    record = Record(configured_test_manager, None)
    with pytest.raises(EmptyCachersError):
        stage_that_does_nothing(record)


def test_cachers_count_mismatch_errors(configured_test_manager):
    """Specifying a different number of cachers than output args should throw a CachersMismatchError."""

    @stage([], ["thing1", "thing2"], [PickleCacher])
    def bad_caching_vibes(record):
        return 1, 2

    record = Record(configured_test_manager, None)
    with pytest.raises(CachersMismatchError):
        bad_caching_vibes(record)


def test_cachers_correct_count(configured_test_manager):
    """Specifying the correct number of cachers for outputs works fine."""

    @stage([], ["thing1", "thing2"], [PickleCacher] * 2)
    def good_caching_vibes(record):
        return 1, 2

    record = Record(configured_test_manager, None)
    good_caching_vibes(record)
    # NOTE: will test caches correctly in caching tests


def test_return_tuple_for_one_output_errors(configured_test_manager):
    """Returning something that is a tuple does not work for a single specified output, this should throw an OutputSignatureError."""

    @stage([], ["my_output"])
    def return_tuple(record):
        multiple_values = (1, 2)
        return multiple_values

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        return_tuple(record)


def test_return_tuplefied_tuple_for_one_output(configured_test_manager):
    """Returning something that is a tuple as a single output value should be tuplefied again."""

    @stage([], ["my_output"])
    def return_tuple(record):
        multiple_values = (1, 2)
        return (multiple_values,)

    record = Record(configured_test_manager, None)
    return_tuple(record)
    assert record.state["my_output"] == (1, 2)


def test_return_array_for_one_output(configured_test_manager):
    """Returning an array as a single output should not have the same issues as a tuple, and should work fine."""

    @stage([], ["my_output"])
    def return_array(record):
        multiple_values = [1, 2]
        return multiple_values

    record = Record(configured_test_manager, None)
    return_array(record)
    assert record.state["my_output"] == [1, 2]


def test_takes_input_from_record(configured_test_manager):
    """A specified input in the stage should pass in that value from the pre-existing record state."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    record.state["my_input"] = 13
    replicate_input(record)
    assert record.state["replicated_input"] == 13


def test_mssing_input_errors(configured_test_manager):
    """A stage input that is not found in the record's state should throw a KeyError."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    with pytest.raises(KeyError):
        replicate_input(record)


def test_missing_input_with_suppress_and_no_default_errors(configured_test_manager):
    """If suppress_missing_inputs is set on the stage but no defaults were specified for the missing ones, we should get an InputSignatureError."""

    @stage(["my_input"], ["replicated_input"], suppress_missing_inputs=True)
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    with pytest.raises(InputSignatureError):
        replicate_input(record)


def test_input_overwritten_by_kwarg(configured_test_manager):
    """A directly passed input argument on the stage function should override whatever was in the record's state previously."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    record.state["my_input"] = 13
    replicate_input(record, my_input=15)
    assert record.state["replicated_input"] == 15


def test_missing_input_but_default_without_suppress_errors(configured_test_manager):
    """Even if there's a default value for an input argument, if suppress isn't specified and we don't pass a value in on the stage call, we should get an error."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_input=15):
        return my_input

    record = Record(configured_test_manager, None)
    with pytest.raises(KeyError):
        replicate_input(record)


def test_missing_input_but_default_with_suppress(configured_test_manager):
    """If a stage has suppressed missing values and we don't pass the missing one in on the stage call, allow the default value to go through without error."""

    @stage(["my_input"], ["replicated_input"], suppress_missing_inputs=True)
    def replicate_input(record, my_input=15):
        return my_input

    record = Record(configured_test_manager, None)
    replicate_input(record)
    assert record.state["replicated_input"] == 15


def test_missing_input_but_kwarg_passed_with_suppress(configured_test_manager):
    """If we specify an input value directly in the stage call, allow it to go through even if it's missing from the record."""

    @stage(["my_input"], ["replicated_input"], suppress_missing_inputs=True)
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    replicate_input(record, my_input=15)
    assert record.state["replicated_input"] == 15


def test_missing_input_but_kwarg_passed_without_suppress(configured_test_manager):
    """If we specify an input value directly in the stage call, allow it to go through even if it's missing from the record and we don't have suppress enabled."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    replicate_input(record, my_input=15)
    assert record.state["replicated_input"] == 15


def test_input_name_incorrect(configured_test_manager):
    """Stage input string names that do not match the function signature variable names should throw an InputSignatureError."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_wrong_input):
        return my_wrong_input

    record = Record(configured_test_manager, None)
    record.state["my_input"] = 13
    with pytest.raises(InputSignatureError):
        replicate_input(record)


def test_lazy_obj_in_record(configured_test_manager):
    """Ensure a lazy object is put into state from a lazy output stage, and ensure that it reloads correctly when used."""

    @stage([], [Lazy("tester")], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    @stage(["tester"], outputs=["output"])
    def use_lazy(record, tester):
        return tester

    record = Record(configured_test_manager, None)
    output_stage(record)
    record.state.resolve = False
    assert type(record.state["tester"]) == Lazy
    record.state.resolve = True

    use_lazy(record)
    assert record.state["output"] == "hello world"


def test_lazy_disabled_on_ignore_lazy(configured_test_manager):
    """Lazy objects should be replaced with str counterparts if ignore_lazy is set on the manager."""

    @stage([], [Lazy("tester")], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    configured_test_manager.ignore_lazy = True
    record = Record(configured_test_manager, None)
    output_stage(record)
    assert type(record.state["tester"]) == str
    record.state.resolve = False
    assert record.state["tester"] == "hello world"


def test_lazy_forced_on_manager_lazy(configured_test_manager):
    """All objects should be lazy if lazy is set on the manager."""

    @stage([], ["tester"], cachers=[PickleCacher])
    def output_stage(record):

        return "hello world"

    configured_test_manager.lazy = True
    record = Record(configured_test_manager, None)
    output_stage(record)
    record.state.resolve = False
    assert type(record.state["tester"]) == Lazy
    record.state.resolve = True
    assert record.state["tester"] == "hello world"


def test_error_on_missing_lazy_cacher(configured_test_manager):
    """If no cacher is given for a lazy object, an error should be thrown."""

    @stage([], [Lazy("tester")])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        output_stage(record)
    # record.state.resolve = False
    # assert type(record.state["tester"]) == Lazy
    # assert type(record.state["tester"].cacher) == PickleCacher


def test_cacher_injected_on_manager_lazy(configured_test_manager):
    """If lazy is set on the manager and an object has no cacher, inject a PickleCacher."""

    @stage([], ["tester"])
    def output_stage(record):
        return "hello world"

    configured_test_manager.lazy = True
    record = Record(configured_test_manager, None)
    output_stage(record)
    record.state.resolve = False
    assert type(record.state["tester"]) == Lazy
    assert type(record.state["tester"].cacher) == PickleCacher


def test_cacher_not_injected_on_manager_ignore_lazy(configured_test_manager):
    """If ignore_lazy is set on the manager and a Lazy object has no cacher, do not error."""
    # NOTE: in theory this behavior is so that if someone gives you buggy stages, you can "fix" the stage without having to actually go in and modify the cachers.

    @stage([], [Lazy("tester")])
    def output_stage(record):
        return "hello world"

    configured_test_manager.ignore_lazy = True
    record = Record(configured_test_manager, None)
    output_stage(record)
    record.state.resolve = False
    assert type(record.state["tester"]) == str
    assert record.state["tester"] == "hello world"


# --------------------------
# @aggregate tests
# --------------------------


def test_aggregate_stores_output_in_record(configured_test_manager):
    """An aggregate output should exist in the record state."""

    @aggregate(["output"])
    def small_aggregate(record, records):
        return "hello world"

    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])  # TODO: blank records array crashes??
    assert record.state["output"] == "hello world"


def test_aggregate_stores_multiple_outputs_in_record(configured_test_manager):
    """Multiple returned outputs from the stage should exist in the record state."""

    @aggregate(["thing1", "thing2"])
    def small_aggregate(record, records):
        return "hello world", 13

    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])
    assert record.state["thing1"] == "hello world"
    assert record.state["thing2"] == 13


def test_aggregate_returns_less_than_expected_errors(configured_test_manager):
    """A function that doesn't return the same number of objects as specified in the aggregate outputs should throw an OutputSignatureError."""

    @aggregate(["test1", "test2"])
    def output_stage(record, records):
        return "hello world"

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        output_stage(record, [record])


def test_aggregate_empty_cachers_array_errors(configured_test_manager):
    """Don't let the user write [] for cachers list, as it always shortcircuits."""
    # NOTE: technically this could probably just be a special case handled by staging.py, but I think specifying [] for cachers is kind of unclear anyway.
    @aggregate([], [])
    def aggregate_stage_that_does_nothing(record, records):
        i_am_very_important = 5 + 3
        del i_am_very_important

    record = Record(configured_test_manager, None)
    with pytest.raises(EmptyCachersError):
        aggregate_stage_that_does_nothing(record, [record])


def test_aggregate_cachers_count_mismatch_errors(configured_test_manager):
    """Specifying a different number of cachers than output args should throw a CachersMismatchError."""

    @aggregate(["thing1", "thing2"], [PickleCacher])
    def bad_caching_vibes(record, records):
        return 1, 2

    record = Record(configured_test_manager, None)
    with pytest.raises(CachersMismatchError):
        bad_caching_vibes(record, [record])


def test_aggregate_cachers_correct_count(configured_test_manager):
    """Specifying the correct number of cachers for outputs works fine."""

    @aggregate(["thing1", "thing2"], [PickleCacher] * 2)
    def good_caching_vibes(record, records):
        return 1, 2

    record = Record(configured_test_manager, None)
    good_caching_vibes(record, [record])
    # NOTE: will test caches correctly in caching tests


def test_aggregate_return_tuple_for_one_output_errors(configured_test_manager):
    """Returning something that is a tuple does not work for a single specified output, this should throw an OutputSignatureError."""

    @aggregate(["my_output"])
    def return_tuple(record, records):
        multiple_values = (1, 2)
        return multiple_values

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        return_tuple(record, [record])


def test_aggregate_return_tuplefied_tuple_for_one_output(configured_test_manager):
    """Returning something that is a tuple as a single output value should be tuplefied again."""

    @aggregate(["my_output"])
    def return_tuple(record, records):
        multiple_values = (1, 2)
        return (multiple_values,)

    record = Record(configured_test_manager, None)
    return_tuple(record, [record])
    assert record.state["my_output"] == (1, 2)


def test_aggregate_return_array_for_one_output(configured_test_manager):
    """Returning an array as a single output should not have the same issues as a tuple, and should work fine."""

    @aggregate(["my_output"])
    def return_array(record, records):
        multiple_values = [1, 2]
        return multiple_values

    record = Record(configured_test_manager, None)
    return_array(record, [record])
    assert record.state["my_output"] == [1, 2]


def test_aggregate_gets_records_states(configured_test_manager):
    """Inside the aggregate, we should be able to get values from the states of the manually passed in records."""

    @aggregate(["output"])
    def small_aggregate(record, records):
        total = 0
        for r in records:
            total += r.state["input"]
        return total

    record1 = Record(configured_test_manager, None)
    record2 = Record(configured_test_manager, None)
    record3 = Record(configured_test_manager, None)

    record1.state["input"] = 3
    record2.state["input"] = 2
    small_aggregate(record3, [record1, record2])
    assert record3.state["output"] == 5


def test_aggregate_gets_no_manager_records(configured_test_manager):
    """Calling an aggregate stage without records and with no records on the manager should not crash."""

    @aggregate(["output"])
    def small_aggregate(record, records):
        total = 0
        for r in records:
            total += r.state["input"]
        return total

    record = Record(configured_test_manager, None, hide=True)
    small_aggregate(record)
    assert record.state["output"] == 0


def test_aggregate_gets_manager_records_states(configured_test_manager):
    """Inside the aggregate, we should be passed all records on the manager by default, if no records explicitly passed."""

    @aggregate(["output"])
    def small_aggregate(record, records):
        total = 0
        for r in records:
            total += r.state["input"]
        return total

    record1 = Record(configured_test_manager, None)
    record2 = Record(configured_test_manager, None)
    record3 = Record(configured_test_manager, None, hide=True)

    record1.state["input"] = 3
    record2.state["input"] = 2

    small_aggregate(record3)
    assert record3.state["output"] == 5


def test_aggregate_lazy_obj_in_record(configured_test_manager):
    """Ensure a lazy object is put into state from a lazy output aggregate stage, and ensure that it reloads correctly when used."""

    @aggregate([Lazy("output")], cachers=[PickleCacher])
    def small_aggregate(record, records):
        total = 0
        for r in records:
            total += r.state["input"]
        return total

    @stage(["output"], outputs=["actual_output"])
    def use_lazy(record, output):
        return output

    record1 = Record(configured_test_manager, None)
    record2 = Record(configured_test_manager, None)
    record3 = Record(configured_test_manager, None, hide=True)

    record1.state["input"] = 3
    record2.state["input"] = 2

    small_aggregate(record3)
    record3.state.resolve = False
    assert type(record3.state["output"]) == Lazy
    record3.state.resolve = True

    use_lazy(record3)
    assert record3.state["actual_output"] == 5


def test_aggregate_lazy_disabled_on_ignore_lazy(configured_test_manager):
    """Lazy objects should be replaced with str counterparts if ignore_lazy is set on the manager."""

    @aggregate([Lazy("output")], cachers=[PickleCacher])
    def small_aggregate(record, records):
        return "hello world"

    configured_test_manager.ignore_lazy = True
    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])
    record.state.resolve = False
    assert type(record.state["output"]) == str
    assert record.state["output"] == "hello world"


def test_aggregate_lazy_forced_on_manager_lazy(configured_test_manager):
    """All objects should be lazy if lazy is set on the manager."""

    @aggregate(["output"], cachers=[PickleCacher])
    def small_aggregate(record, records):
        return "hello world"

    configured_test_manager.lazy = True
    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])
    record.state.resolve = False
    assert type(record.state["output"]) == Lazy
    record.state.resolve = True
    assert record.state["output"] == "hello world"


def test_aggregate_error_on_missing_lazy_cacher(configured_test_manager):
    """If no cacher is given for a lazy object, an error should be thrown."""

    @aggregate([Lazy("output")])
    def small_aggregate(record, records):
        return "hello world"

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        small_aggregate(record, [record])


def test_aggregate_cacher_injected_on_manager_lazy(configured_test_manager):
    """If lazy is set on the manager and an object has no cacher, inject a PickleCacher."""

    @aggregate(["output"])
    def small_aggregate(record, records):
        return "hello world"

    configured_test_manager.lazy = True
    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])
    record.state.resolve = False
    assert type(record.state["output"]) == Lazy
    assert type(record.state["output"].cacher) == PickleCacher


def test_aggregate_cacher_not_injected_on_manager_ignore_lazy(configured_test_manager):
    """If ignore_lazy is set on the manager and a Lazy object has no cacher, do not error."""
    # NOTE: in theory this behavior is so that if someone gives you buggy stages, you can "fix" the stage without having to actually go in and modify the cachers.

    @aggregate([Lazy("output")])
    def small_aggregate(record, records):
        return "hello world"

    configured_test_manager.ignore_lazy = True
    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])
    record.state.resolve = False
    assert type(record.state["output"]) == str
    assert record.state["output"] == "hello world"


def test_aggregate_auto_resolve_lazy_state(configured_test_manager):
    """A lazy cached object in state should auto-resolve when we access it in an aggregate."""

    @stage([], [Lazy("tester")], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    @aggregate(["reproduced_output"])
    def small_aggregate(record, records):
        return records[0].state["tester"]

    record1 = Record(configured_test_manager, None)
    record2 = Record(configured_test_manager, None, hide=True)

    output_stage(record1)
    record1.state.resolve = False
    assert type(record1.state["tester"]) == Lazy
    record1.state.resolve = True

    small_aggregate(record2, [record1])
    assert record2.state["reproduced_output"] == "hello world"



"""These are integration and unit tests for the overall experiment calls."""

import os
import pytest
from pytest_mock import mocker  # noqa: F401 -- flake8 doesn't see it's used as fixture

from curifactory.experiment import run_experiment
from curifactory.manager import ArtifactManager


# TODO: need to test that specifying no params will default to experiment_name


@pytest.mark.parametrize(
    "experiment_params,expected_manager_params",
    [
        (
            dict(experiment_name="test", parameters_list=["params1", "params2"]),
            dict(
                store_entire_run=False,
                dry=False,
                dry_cache=False,
                custom_name=None,
                run_line="experiment test -p params1 -p params2",
                parallel_lock=None,
                parallel_mode=False,
                lazy=False,
                ignore_lazy=False,
            ),
        ),
        (
            dict(experiment_name="test", parameters_list=[]),
            dict(
                store_entire_run=False,
                dry=False,
                dry_cache=False,
                custom_name=None,
                run_line="experiment test",
                parallel_lock=None,
                parallel_mode=False,
                lazy=False,
                ignore_lazy=False,
            ),
        ),
        (
            dict(
                experiment_name="test",
                parameters_list=["params1"],
                cache_dir_override="test/examples/data/superspecialcache",
            ),
            dict(
                store_entire_run=False,
                dry=False,
                dry_cache=False,
                custom_name=None,
                run_line="experiment test -p params1 --cache test/examples/data/superspecialcache",
                parallel_lock=None,
                parallel_mode=False,
                lazy=False,
                ignore_lazy=False,
            ),
        ),
        (
            dict(
                experiment_name="test",
                parameters_list=["params1"],
                store_entire_run=True,
            ),
            dict(
                store_entire_run=True,
                dry=False,
                dry_cache=False,
                custom_name=None,
                run_line="experiment test -p params1 --store-full",
                parallel_lock=None,
                parallel_mode=False,
                lazy=False,
                ignore_lazy=False,
            ),
        ),
        (
            dict(
                experiment_name="test",
                parameters_list=["params1"],
                dry=True,
                dry_cache=True,
            ),
            dict(
                store_entire_run=False,
                dry=True,
                dry_cache=True,
                custom_name=None,
                run_line="experiment test -p params1 --dry --dry-cache",
                parallel_lock=None,
                parallel_mode=False,
                lazy=False,
                ignore_lazy=False,
            ),
        ),
        (
            dict(
                experiment_name="test",
                parameters_list=["params1"],
                custom_name="custom_test",
            ),
            dict(
                store_entire_run=False,
                dry=False,
                dry_cache=False,
                custom_name="custom_test",
                run_line="experiment test -p params1 --name custom_test",
                parallel_lock=None,
                parallel_mode=False,
                lazy=False,
                ignore_lazy=False,
            ),
        ),
        (
            dict(experiment_name="test", parameters_list=["params1"], lazy=True),
            dict(
                store_entire_run=False,
                dry=False,
                dry_cache=False,
                custom_name=None,
                run_line="experiment test -p params1 --lazy",
                parallel_lock=None,
                parallel_mode=False,
                lazy=True,
                ignore_lazy=False,
            ),
        ),
        (
            dict(experiment_name="test", parameters_list=["params1"], ignore_lazy=True),
            dict(
                store_entire_run=False,
                dry=False,
                dry_cache=False,
                custom_name=None,
                run_line="experiment test -p params1 --ignore-lazy",
                parallel_lock=None,
                parallel_mode=False,
                lazy=False,
                ignore_lazy=True,
            ),
        ),
    ],
)
def test_manager_integration(
    mocker,  # noqa: F811 -- mocker has to be passed in as fixture
    experiment_params,
    expected_manager_params,
):
    """Manager initialization calls should have the correct parameters passed to it based on the run_experiment call."""
    mock = mocker.patch.object(ArtifactManager, "__init__", return_value=None)

    try:
        run_experiment(**experiment_params)
    except AttributeError:
        # NOTE: I'm not actually sure a better way around this, all I want to test is that
        # manager was initialized with what I expect
        pass
    mock.assert_called_once_with(
        "test", **expected_manager_params, status_override=None, notes=None
    )


@pytest.mark.parametrize(
    "local_rank,node_rank,expect_parallel",
    [
        (None, None, False),
        (0, None, False),
        (1, None, True),
        (0, 0, False),
        (1, 0, True),
        (0, 1, True),
        (1, 1, True),
    ],
)
def test_rank_manager_integration(
    mocker,  # noqa: F811 -- mocker has to be passed in as fixture
    local_rank,
    node_rank,
    expect_parallel,
    clear_rank_env_vars,
):
    """Experiment should use parallel mode if RANK env vars are set and indicate not rank 0."""
    if local_rank is not None:
        os.environ["LOCAL_RANK"] = str(local_rank)
    if node_rank is not None:
        os.environ["NODE_RANK"] = str(node_rank)

    mock = mocker.patch.object(ArtifactManager, "__init__", return_value=None)
    try:
        run_experiment(experiment_name="test", parameters_list=["params1"])
    except AttributeError:
        # NOTE: I'm not actually sure a better way around this, all I want to test is that
        # manager was initialized with what I expect
        pass
    mock.assert_called_once_with(
        "test",
        store_entire_run=False,
        dry=False,
        dry_cache=False,
        custom_name=None,
        run_line="experiment test -p params1",
        parallel_lock=None,
        parallel_mode=expect_parallel,
        lazy=False,
        ignore_lazy=False,
        status_override=None,
        notes=None,
    )


def test_basic_params_get_loaded():
    """Loading from the two example parameter files (params1 and params2) should load a total of 3 args instances."""
    results, mngr = run_experiment("basic", ["params1", "params2"], dry=True)
    assert len(mngr.experiment_args_file_list) == 2
    total_args_count = 0
    for key in mngr.experiment_args:
        total_args_count += len(mngr.experiment_args[key])
    assert total_args_count == 3
    assert len(mngr.records) == 3


def test_appropriate_store_registry_use_dry(configuration, clear_filesystem):
    """Running an experiment in dry mode should not create a store or parameters registry."""
    results, mngr = run_experiment("basic", ["params1", "params2"], dry=True)
    assert not os.path.exists(f"{configuration['manager_cache_path']}/store.json")
    assert not os.path.exists(
        f"{configuration['manager_cache_path']}/params_registry.json"
    )


def test_appropriate_store_registry_use_wet(configuration, clear_filesystem):
    """Running an experiment NOT in dry mode SHOULD create a store and parameters registry."""
    results, mngr = run_experiment("basic", ["params1", "params2"], dry=False)
    assert os.path.exists(f"{configuration['manager_cache_path']}/store.json")
    assert os.path.exists(f"{configuration['manager_cache_path']}/params_registry.json")


# def test_parallel_calls_correct_ranges(mocker, parallel_count, expected_ranges):
def test_parallel_calls_count_correct(
    mocker,  # noqa: F811 -- mocker has to be passed in as fixture
):
    """Running an experiment with three Args and with --parallel 3 should spawn 3 processes with the correct global index ranges."""
    mock = mocker.patch("multiprocessing.Process")
    mock.return_value.start = lambda: True
    mock.return_value.join = lambda: True
    mock_queue = mocker.patch("multiprocessing.Queue")
    mock_queue.return_value.get = lambda: [None, "success"]
    run_experiment("basic", ["params1", "params2"], parallel=3)
    assert mock.call_args_list[0].kwargs["args"][-9] == ["0-1"]
    assert mock.call_args_list[1].kwargs["args"][-9] == ["1-2"]
    assert mock.call_args_list[2].kwargs["args"][-9] == ["2-3"]
    assert mock.call_count == 3


def test_parallel_calls_count_correct_limits_threads(
    mocker,  # noqa: F811 -- mocker has to be passed in as fixture
):
    """Running an experiment with more processes than args should still only spawn |args| number of processes."""
    mock = mocker.patch("multiprocessing.Process")
    mock.return_value.start = lambda: True
    mock.return_value.join = lambda: True
    mock_queue = mocker.patch("multiprocessing.Queue")
    mock_queue.return_value.get = lambda: [None, "success"]
    run_experiment("basic", ["params1", "params2"], parallel=4)
    assert mock.call_args_list[0].kwargs["args"][-9] == ["0-1"]
    assert mock.call_args_list[1].kwargs["args"][-9] == ["1-2"]
    assert mock.call_args_list[2].kwargs["args"][-9] == ["2-3"]
    assert mock.call_count == 3


# def test_parallel_calls_count_correct_ranges(
#     mocker,  # noqa: F811 -- mocker has to be passed in as fixture
# ):
#     mock = mocker.patch("multiprocessing.Process")
#     mock_queue = mocker.patch("multiprocessing.Queue")
#     mock_queue.return_value.get = lambda: [None, "success"]
#     run_experiment("basic", ["params1", "params2"], parallel=4)
#     print(mock.call_args_list)
#     assert mock.call_count == 3


def test_parallel_overwrite_removed_after_parallel(
    mocker,  # noqa: F811 -- mocker has to be passed in as fixture
):
    """Using overwrite and parallel at the same time should pass overwrite to the subprocs, but not the main final run."""
    mock = mocker.patch("multiprocessing.Process")
    mock.return_value.start = lambda: True
    mock.return_value.join = lambda: True
    mock_queue = mocker.patch("multiprocessing.Queue")
    mock_queue.return_value.get = lambda: [None, "success"]
    results, mngr = run_experiment(
        "basic",
        ["params1", "params2"],
        stage_overwrites=["nonsense"],
        overwrite_override=True,
        parallel=1,
    )

    assert mock.call_args_list[0].kwargs["args"][2]  # overwrite_override
    assert mock.call_args_list[0].kwargs["args"][15] == ["nonsense"]

    # assert mngr.overwrite_stages = [] # NOTE: unclear if this is desired functionality or not
    for record in mngr.records:
        assert not record.args.overwrite


# TODO: do args_names/args_indices get correctly factored into parallel index calls



"""Testing pathing functions on the manager."""

import os
from pytest_mock import mocker  # noqa: F401 -- flake8 doesn't see it's used as fixture

from curifactory import stage, Record
from curifactory.caching import Cacheable, Lazy, PickleCacher


class FakeCacher(Cacheable):
    def __init__(self, path_override=None):
        super().__init__(".fake", path_override=path_override)

    def load(self):
        return None

    def save(self, obj):
        pass


# -----------------------------------------
# Stage/manager.get_path intergration tests
# -----------------------------------------


def test_stage_integration_basic(
    mocker,  # noqa: F811 -- mocker has to be passed in as fixture
    sample_args,
    configured_test_manager,
):
    """The stage should be calling manager's get_path with the correct parameters."""
    mock = mocker.patch.object(
        configured_test_manager, "get_path", return_value="test_path"
    )

    @stage([], ["test_output"], [FakeCacher])
    def do_thing(record):
        return "hello world"

    record = Record(configured_test_manager, sample_args)
    do_thing(record)
    mock.assert_called_once_with("test_output", record, aggregate_records=None)


def test_stage_integration_path_override(
    mocker,  # noqa: F811 -- mocker has to be passed in as fixture
    sample_args,
    configured_test_manager,
):
    """The stage should be calling manager's get_path with the correct parameters when using a path override in a cacher."""
    mock = mocker.patch.object(
        configured_test_manager, "get_path", return_value="test_path"
    )

    @stage([], ["test_output"], [FakeCacher(path_override="test/examples/WHAT")])
    def do_thing(record):
        return "hello world"

    record = Record(configured_test_manager, sample_args)
    do_thing(record)
    mock.assert_called_once_with(
        "test_output", record, base_path="test/examples/WHAT", aggregate_records=None
    )


def test_stage_integration_storefull(
    mocker,  # noqa: F811 -- mocker has to be passed in as fixture
    sample_args,
    configured_test_manager,
):
    """The stage should be calling manager's get_path twice with the correct parameters."""
    configured_test_manager.store_entire_run = True
    mock = mocker.patch.object(
        configured_test_manager, "get_path", return_value="test_path"
    )

    @stage([], ["test_output"], [FakeCacher])
    def do_thing(record):
        return "hello world"

    record = Record(configured_test_manager, sample_args)
    do_thing(record)

    assert len(mock.call_args_list) == 2
    assert mock.call_args_list[0].args == ("test_output", record)
    assert mock.call_args_list[0].kwargs == dict(aggregate_records=None)
    assert mock.call_args_list[1].args == ("test_output", record)
    assert mock.call_args_list[1].kwargs == dict(output=True, aggregate_records=None)


# -----------------------------------------
# get_path unit tests
# -----------------------------------------


def test_get_path_basic(sample_args, configured_test_manager):
    """Calling get_path with a basic set of args should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    path = configured_test_manager.get_path(
        "test_output", record, aggregate_records=None
    )
    assert path == "test/examples/data/cache/test_sample_hash__test_output"


def test_get_path_basic_w_stagename(sample_args, configured_test_manager):
    """Calling get_path with a basic set of args and a valid stage name should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    configured_test_manager.current_stage_name = "somestage"
    path = configured_test_manager.get_path(
        "test_output", record, aggregate_records=None
    )
    assert path == "test/examples/data/cache/test_sample_hash_somestage_test_output"


def test_get_path_path_override(sample_args, configured_test_manager):
    """Calling get_path with a basic set of args and a path override should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    path = configured_test_manager.get_path(
        "test_output", record, base_path="test/examples/WHAT", aggregate_records=None
    )
    assert path == "test/examples/WHAT/test_sample_hash__test_output"


def test_get_path_store_full(sample_args, configured_test_manager):
    """Calling get_path with store-full should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    configured_test_manager.store_entire_run = True
    ts = configured_test_manager.get_str_timestamp()
    path = configured_test_manager.get_path(
        "test_output", record, output=True, aggregate_records=None
    )
    assert (
        path
        == f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/test_sample_hash__test_output"
    )


def test_get_path_custom_name(sample_args, configured_test_manager):
    """Calling get_path when a custom name is in use should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    configured_test_manager.custom_name = "some_custom_name"
    path = configured_test_manager.get_path(
        "test_output", record, aggregate_records=None
    )
    assert path == "test/examples/data/cache/some_custom_name_sample_hash__test_output"


def test_get_path_custom_name_and_store_full(sample_args, configured_test_manager):
    """Calling get_path when a custom name is in use and storefull is called should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    configured_test_manager.store_entire_run = True
    configured_test_manager.custom_name = "some_custom_name"
    ts = configured_test_manager.get_str_timestamp()
    path = configured_test_manager.get_path(
        "test_output", record, output=True, aggregate_records=None
    )
    assert (
        path
        == f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/some_custom_name_sample_hash__test_output"
    )


def test_get_path_no_args(configured_test_manager):
    """Calling get_path with None args will result in None in the output name."""
    record = Record(configured_test_manager, None)
    path = configured_test_manager.get_path(
        "test_output", record, aggregate_records=None
    )
    assert path == "test/examples/data/cache/test_None__test_output"


# -----------------------------------------
# get_run_output_path unit tests
# -----------------------------------------


def test_get_run_output_path(configured_test_manager):
    ts = configured_test_manager.get_str_timestamp()
    path = configured_test_manager.get_run_output_path()
    assert (
        path
        == f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}"
    )


def test_get_run_output_path_w_obj_name(configured_test_manager):
    ts = configured_test_manager.get_str_timestamp()
    path = configured_test_manager.get_run_output_path("testing_file.txt")
    assert (
        path
        == f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/testing_file.txt"
    )
    assert os.path.exists(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}"
    )
    assert not os.path.isdir(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/testing_file.txt"
    )


def test_get_run_output_path_w_subdirs_obj_name(configured_test_manager):
    ts = configured_test_manager.get_str_timestamp()
    path = configured_test_manager.get_run_output_path("subs/testing_file.txt")
    assert (
        path
        == f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/subs/testing_file.txt"
    )
    assert os.path.exists(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/subs"
    )


# -----------------------------------------
# misc
# -----------------------------------------


def test_write_run_env_output(configured_test_manager):
    """Storing a manager run should output the four expected environment info textfiles into the store-full folder."""
    configured_test_manager.store()
    configured_test_manager.write_run_env_output()
    ts = configured_test_manager.get_str_timestamp()
    # NOTE: the run number is 1 because when we store it we actually populate that information (there's just a default value of 0)
    assert os.path.exists(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/requirements.txt"
    )
    assert os.path.exists(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/environment_meta.txt"
    )
    assert os.path.exists(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/run_info.json"
    )


def test_run_line_sanitization_normal(configured_test_manager):
    """Ensure that a normal store-full run-line will result in a correct reproduction line in the run info."""
    configured_test_manager.run_line = "experiment test -p params1 --store-full"
    configured_test_manager.store_entire_run = True
    configured_test_manager.store()
    ts = configured_test_manager.get_str_timestamp()
    assert (
        configured_test_manager.run_info["reproduce"]
        == f"experiment test -p params1 --cache test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts} --dry-cache"
    )


def test_run_line_sanitization_order(configured_test_manager):
    """The placement of the --store-full flag in the initial run line should not impact the reproduction line."""
    configured_test_manager.run_line = (
        "experiment test -p params1 --store-full --parallel 4"
    )
    configured_test_manager.store_entire_run = True
    configured_test_manager.store()
    ts = configured_test_manager.get_str_timestamp()
    assert (
        configured_test_manager.run_info["reproduce"]
        == f"experiment test -p params1 --parallel 4 --cache test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts} --dry-cache"
    )


def test_run_line_sanitization_overwrite(configured_test_manager):
    """The reproduction line should not include overwrite."""
    configured_test_manager.run_line = (
        "experiment test -p params1 --store-full --parallel 4 --overwrite"
    )
    configured_test_manager.store_entire_run = True
    configured_test_manager.store()
    ts = configured_test_manager.get_str_timestamp()
    assert (
        configured_test_manager.run_info["reproduce"]
        == f"experiment test -p params1 --parallel 4 --cache test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts} --dry-cache"
    )


def test_cache_aware_dict_resolve(configured_test_manager):
    """Ensure when resolve is on, lazy objects in a record's state auto load the thing."""

    @stage([], [Lazy("tester")], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    output_stage(record)
    assert record.state["tester"] == "hello world"


def test_cache_aware_dict_no_resolve(configured_test_manager):
    """Ensure when resolve is off, lazy objects in a record's state remain lazy."""

    @stage([], [Lazy("tester")], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    output_stage(record)
    record.state.resolve = False
    assert type(record.state["tester"]) == Lazy
    assert record.state["tester"].name == "tester"
    assert type(record.state["tester"].cacher) == PickleCacher



"""Functions for making a docker image for an experiment run."""

from curifactory import utils


def build_docker(experiment_name, cache_folder, version):
    """Runs the docker build command. This should output the progress
    to the console as it runs.

    Note:
        This assumes your dockerfile is at :code:`docker/dockerfile`.

    Args:
        experiment_name (str): The name of the experiment that was run.
        cache_folder (str): The folder to grab cached objects from.
        version (str): A run number/timestamp.
    """
    cmd_array = [
        "docker",
        "build",
        "-f",
        "docker/dockerfile",
        "--tag",
        f"{experiment_name}:{version}",
        "--tag",
        f"{experiment_name}",
        "--build-arg",
        f"run_folder={cache_folder}",
        # "--build-arg", # NOTE: (01/03/2022) this is not used in the cookiecutter's dockerfile anymore
        # "conda_env_name=test",
        ".",
        "--progress",
        "tty",
    ]  # use plain to see all output
    # cmd_array = ["docker", "build", "-f", "docker/dockerfile", "."]
    utils.run_command(*cmd_array)
